import logging

import socket
from threading import RLock

from lxdbapi import errors, versions
import ssl

try:
    import httplib
except ImportError:
    import http.client as httplib

__all__ = ['TcpAvaticaConnection', 'NetAvaticaConnection']

logger = logging.getLogger(__name__)

SHDRSZ=12
HDRSZ=40
BODY_MAX_SZ=128*1024

class NetAvaticaConnection(object):
    def close(self):
        raise NotImplementedError('Extend NetAvaticaConnection')

    def request(self, body=None):
        raise NotImplementedError('Extend NetAvaticaConnection')

    def checkSerialization(self):
        raise NotImplementedError('Extend NetAvaticaConnection')


class TcpAvaticaConnection(NetAvaticaConnection):
    taggen = 0

    def __init__(self, url, secure, max_retries):
        """Opens a FTP connection to the RPC server."""
        self._close_lock = RLock()
        self.url = url
        self.secure =secure
        self.max_retries = max_retries if max_retries is not None else 3
        self._opened = False
        self._connect()

    def __del__(self):
        self._close()

    def __exit__(self):
        self._close()

    def _connect(self):
        if not self.secure:
            logger.debug("Using TCP")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            self.tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.tcp_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)
            self.tcp_socket.connect((self.url.hostname, self.url.port))
            self.tcpConn = self.tcp_socket
            self._opened = True
        else:
            logger.debug("Using TCP with SSL")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            self.tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.tcp_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)
            self.tcpConn = ssl.wrap_socket(self.tcp_socket)
            self.tcpConn.connect((self.url.hostname, self.url.port))
            self._opened = True

    def _close(self):
        with self._close_lock:
            if self._opened:
                self._opened = False
                self.tcpConn.close()

    def close(self):
        self._close()

    def sendrq(self, body=None):
        if body is None:
            return None
        tag = ++self.taggen
        tag = tag.to_bytes(4, 'little')
        twrap = 1
        twrap = twrap.to_bytes(1, 'little')
        pad = 0
        pad = pad.to_bytes(2, 'little')
        req_sz = len(body)
        n = 0
        while n < req_sz:
            flags = 0
            l = req_sz - n
            if l > BODY_MAX_SZ:
                flags = 1
                l = BODY_MAX_SZ
            logger.debug(f"sendrq: twrap: {twrap} sz: {l}, tag: {tag}, flags: {flags}")
            msg = bytearray()
            msg.extend(l.to_bytes(4, 'little'))
            msg.extend(tag)
            msg.extend(twrap)
            msg.extend(flags.to_bytes(1, 'little'))
            msg.extend(pad)
            msg.extend(body[n:(n+l)])
            self.tcpConn.sendall(msg)
            n += l


    def recvrq(self):
        response = bytearray()
        flags = 1
        while flags % 2 != 0:
            hdr = bytearray()
            while len(hdr) < 12:
                partial = self.tcpConn.recv(12 - len(hdr))
                if partial is None or len(partial) == 0:
                    logger.debug("No hdr")
                    return None
                hdr.extend(partial)
            resp_sz = int.from_bytes(hdr[:4], 'little')
            tag = int.from_bytes(hdr[4:8], 'little')
            twrap = hdr[8]
            flags = hdr[9]
            logger.debug(f"recvrq: twrap: {twrap} sz: {resp_sz}, tag: {tag}, flags: {flags}, hdr: {hdr}")
            if resp_sz < 0:
                logger.warning("negative msg size (sz {}) tag:{}".format(resp_sz, tag))
                raise IOError('IO Error on RPC request on {}. Got negative size [{}]'.format(self.url, resp_sz))
            if resp_sz == 0:
                return response
            pending = resp_sz
            offset = 0
            while pending > 0:
                partial = self.tcpConn.recv(pending)
                if partial is None:
                    raise IOError("No data. Connection closed")
                l = len(partial)
                if l <= 0:
                    raise IOError("No data. Connection closed")
                logger.debug(f"read {l} bytes")
                offset += l
                pending -= l
                response.extend(partial)
        return response

    def request(self, body=None):
        if body is None:
            return None
        self.sendrq(body)
        return self.recvrq()

    def checkSerialization(self):
        raise NotImplementedError('Should use server-info. Cotact support')


# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the JSON-over-HTTP RPC protocol used by Avatica."""

import pprint
import logging

from lxdbapi import errors, responses_pb2, requests_pb2
from lxdbapi.types import native_2_typed_value, typed_value_2_native
from lxdbapi.network_avatica_connection import TcpAvaticaConnection, NetAvaticaConnection
from lxdbapi import common_pb2
from importlib import import_module

from lxdbapi import versions

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser

__all__ = ['AvaticaClient', 'native_2_typed_value', 'typed_value_2_native']

logger = logging.getLogger(__name__)
AVATICA_CLASS_BASE = "org.apache.calcite.avatica.proto"

clivers = f'{versions.MAJOR_VERSION}.{versions.MINOR_VERSION}'


def get_class(kls):
    """Get class given a fully qualified name of a class"""
    parts = kls.split('$')
    class_ = getattr(import_module("lxdbapi.responses_pb2"), parts[1])
    rv = class_()
    return rv


def parse_url(url):
    url = urlparse.urlparse(url)
    if not url.scheme and not url.netloc and url.path:
        netloc = url.path
        if ':' not in netloc:
            netloc = '{}:8765'.format(netloc)
        return urlparse.ParseResult('http', netloc, '/', '', '', '')
    return url


# Defined in phoenix-core/src/main/java/org/apache/phoenix/exception/SQLExceptionCode.java
SQLSTATE_ERROR_CLASSES = [
    ('08', errors.OperationalError),  # Connection Exception
    ('22018', errors.IntegrityError),  # Constraint violatioin.
    ('22', errors.DataError),  # Data Exception
    ('23', errors.IntegrityError),  # Constraint Violation
    ('24', errors.InternalError),  # Invalid Cursor State
    ('25', errors.InternalError),  # Invalid Transaction State
    ('42', errors.ProgrammingError),  # Syntax Error or Access Rule Violation
    ('XLC', errors.OperationalError),  # Execution exceptions
    ('INT', errors.InternalError),  # Phoenix internal error
]

# Relevant properties as defined by https://calcite.apache.org/avatica/docs/client_reference.html
OPEN_CONNECTION_PROPERTIES = (
    'database',
    'schema',
    'databaseName',
    'user',  # User for the database connection
    'password',  # Password for the user
    'parallel',
    'txn_mode',
    'is_kryo',
    'autoCommit',
    'appTimeZone'
)

ONLY_CLIENT_PROPERTIES = ('distribute', 'encoding')


class AvaticaClient(object):
    """Client for Avatica's RPC server.

    This exposes all low-level functionality that the Avatica
    server provides, using the native terminology. You most likely
    do not want to use this class directly, but rather get connect
    to a server using :func:`connect`.
    """
    connection: NetAvaticaConnection = None

    def __init__(self, url, max_retries=None, avatica_con=None, **kwargs):
        """Constructs a new client object.

        :param url:
            URL of an Avatica RPC server.
        """
        self.url = parse_url(url)
        self.max_retries = max_retries if max_retries is not None else 3
        self.avatica_con = avatica_con
        if kwargs is None or 'secure' not in kwargs.keys():
            self.secure = False
        else:
            self.secure = kwargs['secure']

    def set_avatica_con(self, avatica_con):
        self.avatica_con = avatica_con

    def connect(self):
        self.connection = TcpAvaticaConnection(self.url, self.secure, self.max_retries)

    def close(self):
        if self.connection is not None:
            self.connection.close()

    def _apply(self, request_data):
        logger.debug("Sending request\n%s", request_data)
        if self.connection is None:
            raise IOError('No opened Connection')
        request_name = request_data.__class__.__name__

        req_message = common_pb2.WireMessage()
        req_message.name = AVATICA_CLASS_BASE + ".Requests$" + request_data.__class__.__name__
        req_message.wrapped_message = request_data.SerializeToString()

        body = req_message.SerializeToString()
        response_body = self.connection.request(body)
        if response_body is None:
            raise IOError(f"IO Error on [{self.url}] while expecting a response [{request_name}]")
        # deserialize WireMessage
        resp_message = common_pb2.WireMessage()
        resp_message.ParseFromString(response_body)

        logger.debug("Received response\n%s", resp_message)

        if resp_message.name == 'org.apache.calcite.avatica.proto.Responses$ErrorResponse':
            # Decode the error response to provide a full Error Message:

            errorresponse = get_class(resp_message.name)
            errorresponse.ParseFromString(resp_message.wrapped_message)

            for exception in errorresponse.exceptions:
                exceptionmessage = exception.split('\n')[0]
                if 'NoSuchConnectionException' in exceptionmessage:
                    if self.avatica_con is not None:
                        logger.debug("No such connection exception. Trying to reconnect...")
                        self.avatica_con.open()
                        logger.debug("Successfully opened a new connection")
                        self.set_avatica_con(self.avatica_con)
                        logger.debug("Sending request\n%s", pprint.pformat(request_data))
                        request_data.connection_id = self.avatica_con._id
                        request_name = request_data.__class__.__name__

                        req_message = common_pb2.WireMessage()
                        req_message.name = AVATICA_CLASS_BASE + ".Requests$" + request_data.__class__.__name__
                        req_message.wrapped_message = request_data.SerializeToString()

                        body = req_message.SerializeToString()
                        response_body = self.connection.request(body)
                        # deserialize WireMessage

                        resp_message = common_pb2.WireMessage()
                        resp_message.ParseFromString(response_body)

                        logger.debug("Received response\n%s", resp_message)
                    else:
                        raise errors.InterfaceError(errorresponse.error_message, code=errorresponse.error_code,
                                                    sqlstate=errorresponse.sql_state)
                    break
            if resp_message.name == 'org.apache.calcite.avatica.proto.Responses$ErrorResponse':
                # Decode the error response to provide a full Error Message:
                errorresponse = get_class(resp_message.name)
                errorresponse.ParseFromString(resp_message.wrapped_message)
                raise errors.InterfaceError(errorresponse.error_message, code=errorresponse.error_code,
                                            sqlstate=errorresponse.sql_state,
                                            cause=errorresponse.exceptions)

        return resp_message.wrapped_message

        '''# deserialize response
        response = get_class(wire_message.name)
        response.ParseFromString(wire_message.wrapped_message)'''

        '''if type(response) is responses_pb2.ErrorResponse():
            raise errors.InterfaceError(response.error_message, code=response.error_code, sqlstate=response.sql_state,
                                        cause=response.exceptions)
        return response'''

    def get_catalogs(self, connection_id):
        request = requests_pb2.CatalogsRequest()
        request.connection_id = connection_id
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def get_schemas(self, connection_id, catalog=None, schemaPattern=None):
        request = requests_pb2.SchemasRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
        if schemaPattern is not None:
            request.schema_pattern = schemaPattern
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def get_tables(self, connection_id, catalog=None, schemaPattern=None, tableNamePattern=None, typeList=None):
        request = requests_pb2.TablesRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
            request.has_catalog = True
        if schemaPattern is not None:
            request.schema_pattern = schemaPattern
            request.has_schema_pattern = True
        if tableNamePattern is not None:
            request.table_name_pattern = tableNamePattern
            request.has_table_name_pattern = True
        if typeList is not None:
            request.has_type_list = True
            request.type_list.extend(typeList)
        request.has_type_list = typeList is not None
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def get_columns(self, connection_id, catalog=None, schemaPattern=None, tableNamePattern=None,
                    columnNamePattern=None):
        request = requests_pb2.ColumnsRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
        if schemaPattern is not None:
            request.schema_pattern = schemaPattern
        if tableNamePattern is not None:
            request.table_name_pattern = tableNamePattern
        if columnNamePattern is not None:
            request.column_name_pattern = columnNamePattern
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def get_table_types(self, connection_id):
        request = requests_pb2.TableTypesRequest()
        request.connection_id = connection_id
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def get_type_info(self, connection_id):
        request = requests_pb2.TypeInfoRequest()
        request.connection_id = connection_id
        response_data = self._apply(request)
        response = responses_pb2.ResultSetResponse()
        response.ParseFromString(response_data)
        return response

    def connection_sync(self, connection_id, connProps=None):
        """Synchronizes connection properties with the server.

        :param connection_id:
            ID of the current connection.

        :param connProps:
            Dictionary with the properties that should be changed.

        :returns:
            A ``common_pb2.ConnectionProperties`` object.
        """
        if connProps is None:
            connProps = {}

        request = requests_pb2.ConnectionSyncRequest()
        request.connection_id = connection_id
        request.conn_props.auto_commit = connProps.get('autoCommit', False)
        request.conn_props.has_auto_commit = True
        request.conn_props.read_only = connProps.get('readOnly', False)
        request.conn_props.has_read_only = True
        request.conn_props.transaction_isolation = connProps.get('transactionIsolation', 0)
        request.conn_props.catalog = connProps.get('catalog', '')
        request.conn_props.schema = connProps.get('schema', '')

        response_data = self._apply(request)
        response = responses_pb2.ConnectionSyncResponse()
        response.ParseFromString(response_data)
        return response.conn_props

    def open_connection(self, connection_id, info=None):
        """Opens a new connection.

        :param connection_id:
            ID of the connection to open.
        """
        request = requests_pb2.OpenConnectionRequest()
        request.connection_id = connection_id
        if info is not None:
            # Info is a list of repeated pairs, setting a dict directly fails
            request.info["LXJDBCVERS"] = clivers
            for k, v in info.items():
                request.info[k] = v

        response_data = self._apply(request)
        response = responses_pb2.OpenConnectionResponse()
        response.ParseFromString(response_data)

    def close_connection(self, connection_id):
        """Closes a connection.

        :param connection_id:
            ID of the connection to close.
        """
        request = requests_pb2.CloseConnectionRequest()
        request.connection_id = connection_id
        self._apply(request)

    def create_statement(self, connection_id):
        """Creates a new statement.

        :param connection_id:
            ID of the current connection.

        :returns:
            New statement ID.
        """
        request = requests_pb2.CreateStatementRequest()
        request.connection_id = connection_id

        response_data = self._apply(request)
        response = responses_pb2.CreateStatementResponse()
        response.ParseFromString(response_data)
        return response.statement_id

    def close_statement(self, connection_id, statement_id):
        """Closes a statement.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to close.
        """
        request = requests_pb2.CloseStatementRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id

        self._apply(request)

    def prepare_and_execute(self, connection_id, statement_id, sql, max_rows_total=None, first_frame_max_size=None):
        """Prepares and immediately executes a statement.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to prepare.

        :param sql:
            SQL query.

        :param max_rows_total:
            The maximum number of rows that will be allowed for this query.

        :param first_frame_max_size:
            The maximum number of rows that will be returned in the first Frame returned for this query.

        :returns:
            Result set with the signature of the prepared statement and the first frame data.
        """
        request = requests_pb2.PrepareAndExecuteRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        request.sql = sql
        if max_rows_total is not None:
            request.max_rows_total = max_rows_total
        if first_frame_max_size is not None:
            request.first_frame_max_size = first_frame_max_size

        response_data = self._apply(request)
        response = responses_pb2.ExecuteResponse()
        response.ParseFromString(response_data)
        return response.results

    def prepare(self, connection_id, sql, max_rows_total=None):
        """Prepares a statement.

        :param connection_id:
            ID of the current connection.

        :param sql:
            SQL query.

        :param max_rows_total:
            The maximum number of rows that will be allowed for this query.

        :returns:
            Signature of the prepared statement.
        """
        request = requests_pb2.PrepareRequest()
        request.connection_id = connection_id
        request.sql = sql
        if max_rows_total is not None:
            request.max_rows_total = max_rows_total

        response_data = self._apply(request)
        response = responses_pb2.PrepareResponse()
        response.ParseFromString(response_data)
        return response.statement

    def switch(self, type):
        switcher = {
            'INTEGER': ['number_value', 12]

        }
        return switcher.get(type)

    def build_parameter_values(self, parameter_values, signature):
        parameters_typed = []
        for index, param in enumerate(parameter_values):
            param_data_type = signature.parameters[index].type_name
            valueTemp = common_pb2.TypedValue()
            if (self.switch(param_data_type)[0] == 'number_value'):
                valueTemp.number_value = param
            parameters_typed.append(valueTemp)
        return parameters_typed

    def execute(self, connection_id, statement_id, signature, parameter_values=None, first_frame_max_size=None):
        """Returns a frame of rows.

        The frame describes whether there may be another frame. If there is not
        another frame, the current iteration is done when we have finished the
        rows in the this frame.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to fetch rows from.

        :param signature:
            common_pb2.Signature object

        :param parameter_values:
            A list of parameter values, if statement is to be executed; otherwise ``None``.

        :param first_frame_max_size:
            The maximum number of rows that will be returned in the first Frame returned for this query.

        :returns:
            ExecuteResponse
        """
        request = requests_pb2.ExecuteRequest()
        request.statementHandle.id = statement_id
        request.statementHandle.connection_id = connection_id
        request.statementHandle.signature.CopyFrom(signature)
        if parameter_values is not None:
            request.parameter_values.extend(parameter_values)
            request.has_parameter_values = True
        if first_frame_max_size is not None:
            request.first_frame_max_size = first_frame_max_size

        response_data = self._apply(request)
        response = responses_pb2.ExecuteResponse()
        response.ParseFromString(response_data)
        return response

    def writeBlob(self, connection_id, statement_id, blob_values):
        request = requests_pb2.WriteBlobRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        for b in blob_values:
            request.blob_values.append(b)

        response_data = self._apply(request)
        response = responses_pb2.ExecuteResponse()
        response.ParseFromString(response_data)
        return response

    def executeBatch(self, connection_id, statement_id, signature, parameter_values=None, first_frame_max_size=None):
        """Returns a frame of rows

            The frame describes whether there may be another frame. If there is not
            another frame, the current iteration is done when we have finished the
            rows in the this frame.

            :param connection_id:
                ID of the current connection.

            :param statement_id:
                ID of the statement to fetch rows from.

            :param signature:
                common_pb2.Signature object

            :param parameter_values:
                A list of list of parameter values, if statement is to be executed; otherwise ``None``.

            :param first_frame_max_size:
                The maximum number of rows that will be returned in the first Frame returned for this query.

            :returns:
               ExecuteBatchResponse
        """
        request = requests_pb2.ExecuteBatchRequest()
        request.statement_id = statement_id
        request.connection_id = connection_id
        if parameter_values is not None:
            for parameter in parameter_values:
                updateBatch = requests_pb2.UpdateBatch()
                updateBatch.parameter_values.extend(parameter)
                request.updates.extend((updateBatch,))

        response_data = self._apply(request)
        response = responses_pb2.ExecuteBatchResponse()
        response.ParseFromString(response_data)
        return response

    def writeBlobBatch(self, connection_id, statement_id, blob_values):
        request = requests_pb2.WriteBlobBatchRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        for b in blob_values:
            request.blob_values.append(b)

        response_data = self._apply(request)
        response = responses_pb2.ExecuteBatchResponse()
        response.ParseFromString(response_data)
        return response

    def fetch(self, connection_id, statement_id, offset=0, frame_max_size=None):
        """Returns a frame of rows.

        The frame describes whether there may be another frame. If there is not
        another frame, the current iteration is done when we have finished the
        rows in the this frame.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to fetch rows from.

        :param offset:
            Zero-based offset of first row in the requested frame.

        :param frame_max_size:
            Maximum number of rows to return; negative means no limit.

        :returns:
            Frame data, or ``None`` if there are no more.
        """
        request = requests_pb2.FetchRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        request.offset = offset
        if frame_max_size is not None:
            request.frame_max_size = frame_max_size
            # request.frame_max_size = -1

        response_data = self._apply(request)
        response = responses_pb2.FetchResponse()
        response.ParseFromString(response_data)
        return response.frame

    def commit(self, connectionId):
        request = requests_pb2.CommitRequest(connection_id=connectionId)

        response_data = self._apply(request)
        response = responses_pb2.CommitResponse()
        response.ParseFromString(response_data)

    def rollback(self, connectionId):
        request = requests_pb2.RollbackRequest(connection_id=connectionId)

        response_data = self._apply(request)
        response = responses_pb2.RollbackResponse()
        response.ParseFromString(response_data)

    def read_blob(self, connection_id, statement_id, blob_id, offset):
        request = requests_pb2.ReadBlobRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        request.blob_info.id = blob_id
        request.blob_info.offset = offset
        request.blob_info.done = False
        response_data = self._apply(request)
        response = responses_pb2.ReadBlobResponse()
        response.ParseFromString(response_data)
        return response.blob_value

    def serve_lx_action(self, action, params=None):
        request = requests_pb2.LxActionRequest()
        request.action = action
        if params is not None:
            request.params = list()
            for k, v in params.items():
                param = common_pb2.KeyValuePair()
                param.key = k
                param.value = native_2_typed_value(v, None)
                request.params.append(param)
        response = None
        try:
            response_data = self._apply(request)
            if response_data:
                response = responses_pb2.LxActionResponse()
                response.ParseFromString(response_data)
        except BaseException as ex:
            raise ex

        result = {}
        if response and response.params:
            for info in response.params:
                result[info.key] = typed_value_2_native(info.value, None)
        else:
            logger.debug('No response for LxAction')

        return result





# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'Warning', 'Error', 'InterfaceError', 'DatabaseError', 'DataError',
    'OperationalError', 'IntegrityError', 'InternalError',
    'ProgrammingError', 'NotSupportedError',
]

try:
    _StandardError = StandardError
except NameError:
    _StandardError = Exception


class Warning(_StandardError):
    """Not used by this package, only defined for compatibility
    with DB API 2.0."""


class Error(_StandardError):
    """Exception that is the base class of all other error exceptions.
    You can use this to catch all errors with one single except statement."""

    def __init__(self, message, code=None, sqlstate=None, cause=None):
        super(_StandardError, self).__init__(message, code, sqlstate, cause)

    @property
    def message(self):
        return self.args[0]

    @property
    def code(self):
        return self.args[1]

    @property
    def sqlstate(self):
        return self.args[2]

    @property
    def cause(self):
        return self.args[3]


class InterfaceError(Error):
    """Exception raised for errors that are related to the database
    interface rather than the database itself."""


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""


class DataError(DatabaseError):
    """Exception raised for errors that are due to problems with the
    processed data like division by zero, numeric value out of range,
    etc."""


class OperationalError(DatabaseError):
    """Raised for errors that are related to the database's operation and not
    necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction could
    not be processed, a memory allocation error occurred during
    processing, etc."""


class IntegrityError(DatabaseError):
    """Raised when the relational integrity of the database is affected, e.g. a foreign key check fails."""


class InternalError(DatabaseError):
    """Raised when the database encounters an internal problem."""


class ProgrammingError(DatabaseError):
    """Raises for programming errors, e.g. table not found, syntax error, etc."""


class NotSupportedError(DatabaseError):
    """Raised when using an API that is not supported by the database."""



# Copyright 2017 Dimitri Capitaine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib.parse

from sqlalchemy.connectors.pyodbc import PyODBCConnector
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.sql.compiler import DDLCompiler, SQLCompiler
from sqlalchemy.sql import compiler
from sqlalchemy.engine import default

from sqlalchemy import types, BigInteger, Float
from sqlalchemy.types import INTEGER, BIGINT, SMALLINT, VARCHAR, CHAR, \
    FLOAT, REAL, DATE, BOOLEAN, DECIMAL, TIMESTAMP, TIME, VARBINARY
# TO DO: LXARRAY?

class LeanxcaleCompiler(SQLCompiler):
    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleDDLCompiler(DDLCompiler):

    def visit_primary_key_constraint(self, constraint):
        # if constraint.name is None:
        #    raise CompileError("can't create primary key without a name")
        return DDLCompiler.visit_primary_key_constraint(self, constraint)

    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleExecutionContext(default.DefaultExecutionContext):

    def should_autocommit_text(self, statement):
        pass

    def create_server_side_cursor(self):
        pass

    def fire_sequence(self, seq, type_):
        return self._execute_scalar(
            (
                    "SELECT NEXT VALUE FOR %s"
                    % seq.name
            ),
            type_
        )


class LeanXcaleIdentifierPreparer(compiler.IdentifierPreparer):

    def __init__(self, dialect, server_ansiquotes=False, **kw):

        quote = '"'

        super(LeanXcaleIdentifierPreparer, self).__init__(
            dialect, initial_quote=quote, escape_quote=quote
        )

    def _quote_free_identifiers(self, *ids):
        """Unilaterally identifier-quote any number of strings."""

        return tuple([self.quote_identifier(i) for i in ids if i is not None])

class LeanXcaleGenericTypeCompiler(compiler.GenericTypeCompiler):
    def visit_ARRAY(self, type_, **kw):
        if type_.item_type.python_type == int:
            return "BIGINT ARRAY"
        elif type_.item_type.python_type == float:
            return "DOUBLE ARRAY"
        elif type_.item_type.python_type == str:
            return "VARCHAR ARRAY"
        else:
            raise Exception("ARRAY of type {} is not supported".format(str(type_)))


class LeanxcaleDialect(PyODBCConnector, DefaultDialect):

    name = "leanxcale"

    pyodbc_driver_name = "LeanXcaleODBCDriver"

    ddl_compiler = LeanxcaleDDLCompiler
    preparer = LeanXcaleIdentifierPreparer
    statement_compiler = LeanxcaleCompiler
    type_compiler = LeanXcaleGenericTypeCompiler

    supports_sequences = True
    sequences_optional = True
    supports_multivalues_insert = True

    execution_ctx_cls = LeanxcaleExecutionContext

    preexecute_autoincrement_sequences = True

    @classmethod
    def import_dbapi(cls):
        import lxpyodbc as pyodbc

        # pyodbc.pooling = (
        #     False  # required for Access databases with ODBC linked tables
        # )
        return pyodbc

    @classmethod
    def dbapi(cls):
        import lxpyodbc as pyodbc
        # pyodbc.pooling = (
        #     False  # required for Access databases with ODBC linked tables
        # )
        return pyodbc

    def on_connect(self):
        def do_on_connect(c):
            import lxpyodbc as pyodbc
            c.maxwrite = 1024 * 1024 * 1024
            c.setdecoding(pyodbc.SQL_CHAR, encoding='utf-8')
            c.setdecoding(pyodbc.SQL_WCHAR, encoding='utf-8')
            c.setencoding(encoding='utf-8')

        return do_on_connect

    def do_rollback(self, dbconn):
        dbconn.rollback()

    def do_commit(self, dbconn):
        dbconn.commit()

    def has_sequence(self, connection, sequence_name, schema=None, **kw):
        return sequence_name in self.get_sequence_names(connection, schema)

    def get_sequence_names(self, connection, schema=None, **kw):
        schema = schema if schema is not None else self.default_schema_name
        params = []
        sql = "SELECT DISTINCT(table_name) from LXSYS.TABLES where table_schem <> 'LXSYS'"
        sql += " AND table_type = 'SEQUENCE'"
        if schema:
            sql += " AND table_schem = '?'"
            params.append(str(schema))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(row[0])
            row = cursor.fetchone()
        return result

    def has_table(self, connection, table_name, schema=None, **kw):
        return table_name in self.get_table_names(connection, schema)

    def _get_default_schema_name(self, connection):
        sql = "select table_schem from LXSYS.CONNECTIONS where current"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql)
        row = cursor.fetchone()
        if row:
            return row[0]
        return None

    def get_schema_names(self, connection, **kw):
        sql = "select distinct(table_schem) from LXSYS.TABLES where table_schem <> 'LXSYS'"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql)
        result = []
        schema = cursor.fetchone()
        while schema is not None:
            result.append(str(schema[0]))
            schema = cursor.fetchone()
        return result

    def get_table_names(self, connection, schema=None, **kw):
        schema = schema if schema is not None else self.default_schema_name
        params = []
        sql = "SELECT DISTINCT(table_name) from LXSYS.TABLES where table_schem <> 'LXSYS'"
        sql += " AND table_type IN ('TABLE', 'MATERIALIZED QUERY TABLE')"
        if schema:
            sql += " AND table_schem = ?"
            params.append(str(schema))
        if 'tableNamePattern' in kw.keys():
            tableNamePattern = kw.get('tableNamePattern')
            sql += " AND table_name LIKE ?"
            params.append(str(tableNamePattern))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(str(row[0]))
            row = cursor.fetchone()
        return result

# table_cat   table_schem   table_name   column_name   data_type   type_name   column_size   buffer_length   decimal_digits   num_prec_radix   nullable   remarks   column_def   sql_data_type   sql_datetime_sub   char_octet_length   ordinal_position   is_nullable   scope_cat   scope_schem   scope_table   source_data_type   is_autoincrement   is_generated_column  

    def get_columns(self, connection, table_name=None, schema=None, catalog=None, column_name=None, **kw):
        params = []
        sql = "select column_name, data_type, type_name, nullable, column_def from LXSYS.COLUMNS WHERE 1 = 1"
        schema = schema if schema is not None else self.default_schema_name
        if catalog:
            sql += " AND table_cat = ?"
            params.append(str(catalog))
        if schema:
            sql += "AND table_schem = ?"
            params.append(str(schema))
        if table_name:
            sql += "AND table_name = ?"
            params.append(str(table_name))
        if column_name:
            sql += "AND column_name = ?"
            params.append(str(column_name))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            col_d = {}
            col_d.update({'name': row[0]})
            if row[1] == ARRAY_CODE:	## TODO
                if row[2].split()[0] == 'BIGINT':
                    col_d.update({'item_type': BigInteger})
                elif row[2].split()[0] == 'DOUBLE':
                    col_d.update({'item_type': Float})
                elif row[2].split()[0] == 'VARCHAR':
                    col_d.update({'item_type': VARCHAR})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == ARYI_CODE:	## TODO
                col_d.update({'item_type': BigInteger})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == ARYF_CODE:	## TODO
                col_d.update({'item_type': Float})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == ARYS_CODE:	## TODO
                col_d.update({'item_type': VARCHAR})
                col_d.update({'type': ARRAY(row[5])})
            else:
                col_d.update({'type': COLUMN_DATA_TYPE[row[1]]})
            col_d.update({'nullable': row[3] == 1 if True else False})
            col_d.update({'default': row[4]})
            if row[0] != '_UUID_':
                result.append(col_d)
            row = cursor.fetchone()

        return result

    def get_view_names(self, connection, schema=None, **kw):
        params = []
        schema = schema if schema is not None else self.default_schema_name
        sql = "SELECT DISTINCT(table_name) from LXSYS.TABLES where table_schem <> 'LXSYS'"
        sql += " AND table_type = 'VIEW'"
        if schema:
            sql += " AND table_schem = ?"
            params.append(str(schema))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(row[0])
            row = cursor.fetchone()
        return result

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT column_name FROM LXSYS.PRIMARY_KEYS WHERE table_name = ? ORDER BY key_seq"
        params.append(str(table_name))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        cols = []
        row = cursor.fetchone()
        while row is not None:
            if row[0] != '_UUID_':
                cols.append(row[0])
            row = cursor.fetchone()

        result = {
            'constrained_columns': cols,
        }
        return result
 
# table_cat   table_schem   table_name   column_name   key_name   key_seq   ptable_cat   ptable_schem   ptable_name   pcolumn_name   enable   validate   on_delete  
#  TABLE_CAT   TABLE_SCHEM   TABLE_NAME   COLUMN_NAME   DATA_TYPE   TYPE_NAME   COLUMN_SIZE   BUFFER_LENGTH   DECIMAL_DIGITS   NUM_PREC_RADIX   NULLABLE   REMARKS   COLUMN_DEF   SQL_DATA_TYPE   SQL_DATETIME_SUB   CHAR_OCTET_LENGTH   ORDINAL_POSITION   IS_NULLABLE   SCOPE_CATALOG   SCOPE_SCHEMA   SCOPE_TABLE   SOURCE_DATA_TYPE   IS_AUTOINCREMENT   IS_GENERATEDCOLUMN  

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT table_schem, table_name, key_name, ptable_schem, ptable_name, pcolumn_name, column_name, key_seq"\
              " from LXSYS.FOREIGN_KEYS" \
              " WHERE table_name = ?"
        params.append(str(table_name))
        if schema:
            sql += " and table_schem = ?"
            params.append(str(schema))
        sql += " order by table_schem, table_name, key_name, key_seq"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        fksch = None
        fktbl = None
        fkname = None
        pksch = None
        pktbl = None
        fkcols = None
        pkcols = None
        row = cursor.fetchone()
        while row is not None:
            if fkname != row[2] or fktbl != row[1] or fksch != row[0]:
                if fkname:
                    fk_info = {
                        "constrained_columns": fkcols,
                        "referred_schema": pksch,
                        "referred_table": pktbl,
                        "referred_columns": pkcols,
                        "name": fkname,
                    }
                    result.append(fk_info)
                fksch = row[0]
                fktbl = row[1]
                fkname = row[2]
                pksch = row[3]
                pktbl = row[4]
                fkcols = []
                pkcols = []
            pkcols.append(row[5])
            fkcols.append(row[6])
            row = cursor.fetchone()

        if fkname:
            fk_info = {
                "constrained_columns": fkcols,
                "referred_schema": pksch,
                "referred_table": pktbl,
                "referred_columns": pkcols,
                "name": fkname,
            }
            result.append(fk_info)
        return result

#  non_unique   index_qualifier   index_name   type   ordinal_position   column_name   asc_or_desc   cardinality   pages   filter_condition   table_cat   table_schem   table_name  

    def get_indexes(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT index_name, column_name, cast(non_unique as INTEGER) FROM LXSYS.INDEX_COLUMNS WHERE table_name = ?"
        params.append(str(table_name))
        if schema:
            sql += " AND table_schem = ?"
            params.append(str(schema))
        sql += " order by index_name, ordinal_position"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        name = None
        cols = None
        nonUnique = None
        row = cursor.fetchone()
        while row is not None:
            if name != row[0]:
                if name is not None:
                    item = {
                        'name': name,
                        'column_names': cols,
                        'unique': not nonUnique
                    }
                    result.append(item)
                name = row[0]
                cols = []
                if row[2] is None or int(row[2]) == 0:
                    nonUnique = False
                else:
                    nonUnique = True
            cols.append(row[1])
            row = cursor.fetchone()

        if name is not None:
            item = {
                'name': name,
                'column_names': cols,
                'unique': not nonUnique
            }
            result.append(item)
        return result

class BOOLEAN(types.Boolean):
    __visit_name__ = "BOOLEAN"

class TINYINT(types.Integer):
    __visit_name__ = "INTEGER"

class UTINYINT(types.Integer):
    __visit_name__ = "INTEGER"

class UINTEGER(types.Integer):
    __visit_name__ = "INTEGER"

class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"

class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"

class UDOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"

class UFLOAT(types.FLOAT):
    __visit_name__ = "FLOAT"

class ULONG(types.BIGINT):
    __visit_name__ = "BIGINT"

class UTIME(types.TIME):
    __visit_name__ = "TIME"

class UDATE(types.DATE):
    __visit_name__ = "DATE"

class UTIMESTAMP(types.TIMESTAMP):
    __visit_name__ = "TIMESTAMP"

class ROWID(types.String):
    __visit_name__ = "VARCHAR"

class ARRAY(types.ARRAY):	## TODO
    __visit_name__ = "ARRAY"
    def __init__(self, type):
        types.ARRAY.__init__(self, type)

class CIDR(types.String):
    __visit_name__ = "CIDR"

class REAL(types.REAL):
    __visit_name__ = "REAL"

# CIDR, TIMESTAMP_LTZ

# lx=specific codes
CIDR_CODE = 3000
TSLTZ_CODE = 3001
ARYI_CODE = 2104
ARYF_CODE = 2108
ARYS_CODE= 2112
TSTZ_CODE = 2014
ARRAY_CODE = 2003
 
# 16: BOOLEAN,
# -6: TINYINT,
# 5: SMALLINT,
# 4: INTEGER,
# -5: BIGINT,
# 3: DECIMAL,
# 6: FLOAT,
# 7: REAL,
# 8: DOUBLE,
# 91: DATE,
# 92: TIME,
# 2014: TIMESTAMP_TZ,
# 93: TIMESTAMP,
# 3001: TIMESTAMP_LTZ,
# #1111: INTERVAL,
# 1: CHAR,
# 12: VARCHAR,
# 2005: CLOB,
# -2: BINARY,
# -3: VARBINARY,
# 2004: BLOB,
# 2003: MULTISET,
# 2003: ARRAY,
# 1111: MAP,
# 2001: DISTINCT,
# 2002: STRUCTURED,
# 2002: ROW,
# 1111: OTHER,
# 2012: CURSOR,
# 1113: COLUMN_LIST,
# 2015: GEOMETRY,
# 3000: CIDR,

COLUMN_DATA_TYPE = {
    -6: TINYINT,
    -5: BIGINT,
    -3: VARBINARY,
    1: CHAR,
    3: DECIMAL,
    4: INTEGER,
    5: SMALLINT,
    6: FLOAT,
    7: REAL,
    8: DOUBLE,
    9: UINTEGER,
    10: ULONG,
    11: UTINYINT,
    12: VARCHAR,
    13: ROWID,
    14: UFLOAT,
    15: UDOUBLE,
    16: BOOLEAN,
    18: UTIME,
    19: UDATE,
    20: UTIMESTAMP,
    91: DATE,
    92: TIME,
    93: TIMESTAMP,
#  ARYI_CODE
# ARYF_CODE
# ARYS_CODE
    TSTZ_CODE: TIMESTAMP,
    CIDR_CODE: CIDR,
    TSLTZ_CODE: TIMESTAMP,
    ## TODO array types
}



import logging
import os
import time
from threading import RLock, Thread

__all__ = ['MultiConnectionI', 'ServiceConnectionI', 'ConnectionManager', 'RefreshableConnectionManager']

logger = logging.getLogger(__name__)


class MultiConnectionI(object):

    def id(self):
        raise NotImplementedError('Extend ConnectionManager')

    def reconnect(self, addr=None):
        raise NotImplementedError('Extend ConnectionManager')

    def qe_address(self):
        raise NotImplementedError('Extend ConnectionManager')

    def unset_do_reconnect(self):
        raise NotImplementedError('Extend ConnectionManager')

    def set_do_reconnect(self):
        raise NotImplementedError('Extend ConnectionManager')


class ServiceConnectionI(object):

    def qe_addresses(self):
        raise NotImplementedError('Extend ConnectionManager')


class ConnectionManager(object):

    def add_connection(self, conn: MultiConnectionI):
        raise NotImplementedError('Extend ConnectionManager')

    def remove_connection(self, conn: MultiConnectionI):
        raise NotImplementedError('Extend ConnectionManager')

    def reconnect(self, conn: MultiConnectionI):
        raise NotImplementedError('Extend ConnectionManager')

    def update(self, addresses: set):
        raise NotImplementedError('Extend ConnectionManager')

    def next_qe(self):
        return None

    def invalidate_address(self, addr: str):
        raise NotImplementedError('Extend ConnectionManager')

    def validate_address(self, addr: str):
        raise NotImplementedError('Extend ConnectionManager')

    def is_valid_address(self, addr: str):
        return True

    def is_stopped(self):
        raise NotImplementedError('Extend ConnectionManager')

    def stop(self):
        raise NotImplementedError('Extend ConnectionManager')

    def info(self):
        return None


class RefreshableConnectionManager(ConnectionManager):

    def __init__(self, service_connection: ServiceConnectionI):
        self._stopped = False
        self._stopped_lock = RLock()
        self._service_connection = service_connection
        self._thread = Thread(target=refresh_conn_mgr, args=[self], daemon=True)
        self._thread.start()

    def add_connection(self, conn: MultiConnectionI):
        super().add_connection(conn)

    def remove_connection(self, conn: MultiConnectionI):
        super().remove_connection(conn)

    def update(self, addresses: set):
        super().update(*addresses)

    def next_qe(self):
        return super().next_qe()

    def invalidate_address(self, addr: str):
        super().invalidate_address(addr)

    def validate_address(self, addr: str):
        super().validate_address(addr)

    def is_valid_address(self, addr: str):
        return super().is_valid_address(addr)

    def reconnect_address(self, addr: str, connection: MultiConnectionI):
        is_reconnected = False
        logger.debug("Reconnecting connection {} to QE {}".format(connection.id(), addr))
        try:
            connection.reconnect(addr)
            # self.add_connection(connection)
            is_reconnected = True
        except Exception as ex:
            msg = str(ex)
            if 'Connection already exists' in msg:
                logger.debug("Connection {} already exists in {}".format(connection.id(), addr))
                logger.debug("Connection {} already exists in {}: {}".format(connection.id(), addr, ex))
                is_reconnected = True
            else:
                logger.error("Cannot reconnect {} to {}: {}".format(connection.id(), addr, ex))
        return is_reconnected

    def reconnect_retries(self):
        raise NotImplementedError('Extend RefreshableConnectionManager')

    def get_service_connection(self):
        return self._service_connection

    def is_stopped(self):
        with self._stopped_lock:
            result = self._stopped
        return result

    def stop(self):
        with self._stopped_lock:
            self._stopped = True

    def reconnect(self, connection: MultiConnectionI):
        addr = connection.qe_address()
        is_reconnected = False
        retries = self.reconnect_retries()
        self.remove_connection(connection)
        while is_reconnected is False and retries >= 0:
            retries -= 1
            new_addr = self.next_qe()
            if new_addr is None:
                logger.error("No other address but {}. Retry once".format(addr))
                retries = -1
                new_addr = addr

            if self.reconnect_address(new_addr, connection):
                is_reconnected = True
                connection.unset_do_reconnect()
            else:
                self.invalidate_address(new_addr)
                logger.error("Failed reconnection from {} to {}".format(addr, new_addr))
        return is_reconnected

    def info(self):
        with self._stopped_lock:
            result = 'is stopped: ' + str(self._stopped)
        return result


def refresh_conn_mgr(manager: RefreshableConnectionManager):
    period = int(os.environ.get("LX_CM_PERIOD", 1))
    while not manager.is_stopped():
        try:
            service = manager.get_service_connection()
            qe_addresses = None
            if service is not None:
                addrs = service.qe_addresses()
                if addrs is not None:
                    logger.debug("Qe addresses: {}".format(addrs))
                    qe_addresses = set(addrs)
            if qe_addresses is None:
                logger.warning("No available Qe addresses found")
            else:
                manager.update(qe_addresses)
        except BaseException as ex:
            logger.error("server refresh failed: {}".format(ex))
            manager.stop()
            raise ex
        time.sleep(period)



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import base64
import datetime
from codecs import IncrementalEncoder, CodecInfo, BufferedIncrementalEncoder
from decimal import Decimal
from io import RawIOBase, TextIOBase, UnsupportedOperation, IOBase

from lxdbapi.common_pb2 import Rep, TypedValue
from lxdbapi.errors import ArrayTypeException
from lxdbapi import common_pb2

__all__ = [
	'Date', 'Time', 'Timestamp', 'DateFromTicks', 'TimeFromTicks', 'TimestampFromTicks',
	'Binary', 'STRING', 'BINARY', 'NUMBER', 'DATETIME', 'ROWID', 'BOOLEAN', 'javaTypetoNative',
	'nativeToParamType', 'LXARRAY', 'typed_value_2_native', 'native_2_typed_value', 'LxRawReader', 'LxTextReader'
]


def Date(year, month, day):
	"""Constructs an object holding a date value."""
	return datetime.date(year, month, day)


def Time(hour, minute, second):
	"""Constructs an object holding a time value."""
	return datetime.time(hour, minute, second)


def Timestamp(year, month, day, hour, minute, second):
	"""Constructs an object holding a datetime/timestamp value."""
	return datetime.datetime(year, month, day, hour, minute, second)


def DateFromTicks(ticks):
	"""Constructs an object holding a date value from the given UNIX timestamp."""
	return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
	"""Constructs an object holding a time value from the given UNIX timestamp."""
	return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
	"""Constructs an object holding a datetime/timestamp value from the given UNIX timestamp."""
	return Timestamp(*time.localtime(ticks)[:6])


def Binary(value):
	"""Constructs an object capable of holding a binary (long) string value."""
	if isinstance(value, _BinaryString):
		return value
	return _BinaryString(base64.b64encode(value))


def time_from_java_sql_time(n):
	dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)
	return dt.time()


def time_to_java_sql_time(t):
	return int(((t.hour * 60 + t.minute) * 60 + t.second) * 1000 + t.microsecond / 1000)


def date_from_java_sql_date(n):
	return datetime.date(1970, 1, 1) + datetime.timedelta(days=n)


def date_to_java_sql_date(d):
	if isinstance(d, datetime.datetime):
		d = d.date()
	td = d - datetime.date(1970, 1, 1)
	return td.days


def datetime_from_java_sql_timestamp(n):
	return datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc) + datetime.timedelta(milliseconds=n)


def datetime_to_java_sql_timestamp(d):
	td = d - datetime.datetime(1970, 1, 1).replace(tzinfo=datetime.timezone.utc)
	return int(td.microseconds / 1000 + (td.seconds + td.days * 24 * 3600) * 1000)


class _BinaryString(str):
	pass


class ColumnType(object):

	def __init__(self, eq_types):
		self.eq_types = tuple(eq_types)
		self.eq_types_set = set(eq_types)

	def __cmp__(self, other):
		if other in self.eq_types_set:
			return 0
		if other < self.eq_types:
			return 1
		else:
			return -1


STRING = ColumnType(['VARCHAR', 'CHAR'])
"""Type object that can be used to describe string-based columns."""

BINARY = ColumnType(['BINARY', 'VARBINARY'])
"""Type object that can be used to describe (long) binary columns."""

NUMBER = ColumnType(['INTEGER', 'UNSIGNED_INT', 'BIGINT', 'UNSIGNED_LONG', 'TINYINT', 'UNSIGNED_TINYINT', 'SMALLINT',
					 'UNSIGNED_SMALLINT', 'FLOAT', 'UNSIGNED_FLOAT', 'DOUBLE', 'UNSIGNED_DOUBLE', 'DECIMAL'])
"""Type object that can be used to describe numeric columns."""

DATETIME = ColumnType(['TIME', 'DATE', 'TIMESTAMP', 'UNSIGNED_TIME', 'UNSIGNED_DATE', 'UNSIGNED_TIMESTAMP'])
"""Type object that can be used to describe date/time columns."""

ROWID = ColumnType([])
"""Only implemented for DB API 2.0 compatibility, not used."""

BOOLEAN = ColumnType(['BOOLEAN'])
"""Type object that can be used to describe boolean columns. This is a lxdbapi-specific extension."""


# XXX ARRAY
class LXARRAY():
	__value = []
	__type = None

	def __init__(self, value, type):
		self.__value = value
		self.__type = type

	def __str__(self):
		return f'({self.__type}){self.__value}'

	def getvalue(self):
		return self.__value

	def setvalue(self, value):
		self.__value = value

	def gettype(self):
		return self.__type

	def settype(self, type):
		self.__type = type


BUF_SZ = 65536


class LxRawReader(RawIOBase):
	"""
	API to read bytes from any object.
	Need to provide the read and close functions that performs the action over the byte source object.

	The constructor has the following arguments:
		- obj: the object where to read the bytes from
		- readf: function that will read bytes from obj into a pre-allocated, writable bytes-like object. Its signature
		 should be:
			+ <fname>(buf:WritableBuffer, offset, obj, len) where :
				- buf is a WritableBuffer object (probably a memoryview) where to write the bytes read from obj
				- obj is the object to read bytes from
				- It return the number of bytes read. None when eof is reached or no more bytes are available.
		- closef: function that will close obj when needed. Could be None when not needed. Its signature should be:
			+ <fname>(obj) where obj is the object to read bytes from

	Note that this API won't allow reading all the bytes at once.
	"""

	def __init__(self, obj, readf=None, closef=None):
		self.obj = obj
		if readf is None:
			self.readf = lambda buf, o: o.readinto(buf)
		else:
			self.readf = readf
		if closef is None:
			self.closef = lambda o: o.close()
		else:
			self.closef = closef
		self._closed = False
		self.offset = 0

	def __enter__(self):
		return self

	def __del__(self):
		self.close()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.close()

	def close(self) -> None:
		if not self._closed:
			self._closed = True
			if self.closef is not None:
				self.closef(self.obj)

	@property
	def closed(self):
		return self._closed

	def read(self, __size=...):
		if self._closed:
			return None
		if __size is None:
			raise RuntimeError("Read all not allowed. Use size argument")
		if __size < 0:
			raise RuntimeError("Read all not allowed. Use size argument")
		if __size == 0:
			return bytes(0)
		if __size > BUF_SZ:
			raise RuntimeError(f"Can read at most {BUF_SZ} bytes")
		offset = 0
		buf = bytearray(__size)
		mv = memoryview(buf)
		while offset < __size:
			n = self.readf(mv[offset:], self.obj)
			if n is None or n <= 0:
				self.close()
				break
			offset += n
		self.offset += offset
		if offset == 0:
			return None
		if offset == __size:
			return bytes(buf)
		return bytes(buf[:offset])

	def readall(self):
		raise RuntimeError("Read all not allowed")

	def readinto(self, buf):
		if self._closed:
			return None
		if buf is None:
			return 0
		n = self.readf(buf, self.obj)
		self.offset += n
		return n

	def fileno(self):
		raise OSError("No fileno")

	def readlines(self, __hint=...):
		raise OSError("Not allowed. Use readinto")

	def readline(self, __size=...) -> bytes:
		raise OSError("Not allowed. Use readinto")

	def isatty(self) -> bool:
		return False

	def flush(self) -> None:
		return None

	def readable(self) -> bool:
		return True

	def seekable(self) -> bool:
		return False

	def seek(self, __offset: int, __whence=...):
		raise OSError("Not seekable")

	def tell(self) -> int:
		raise OSError("Not seekable")

	def truncate(self, __size=...):
		raise OSError("Not seekable")

	def writable(self) -> bool:
		return False

	def write(self, __b):
		raise OSError("Not writeable")

	def writelines(self, __lines):
		raise OSError("Not writeable")


class LxText2Bin(object):
	def __init__(self, io: TextIOBase, codeci: CodecInfo):
		self.io = io
		self.iencoder = codeci.incrementalencoder()
		self.pending = None
		self.current = ""
		self.done = False
		self._next()
		self.first = self.current

	def _next(self):
		self.current = ""
		offset = 0
		while offset < BUF_SZ:
			s = self.io.read(BUF_SZ - offset)
			if s is None:
				self.done = True
				break
			l = len(s)
			if l <= 0:
				self.done = True
				break
			self.current += s
			offset += l

	def close(self):
		self.io.close()
		self.io = None

	def _pendingreadinto(self, buf, offset, l):
		n = 0
		if self.pending is not None:
			sz = len(self.pending)
			if sz > l:
				last = offset + l
				buf[offset:last] = self.pending[0:l]
				self.pending = self.pending[l:]
				n += l
				return n
			last = offset + sz
			buf[offset:last] = self.pending[0:]
			self.pending = None
			n += sz
		return n

	def readinto(self, buf):
		if buf is None:
			return 0
		sz = len(buf)
		if sz == 0:
			return 0
		n = self._pendingreadinto(buf, 0, sz)

		while n < sz:
			slen = len(self.current)
			if slen <= 0:
				if self.done:
					break
				self._next()
			b = self.iencoder.encode(self.current, self.done)
			self._next()
			self.pending = b
			l = sz - n
			nn = self._pendingreadinto(buf, n, l)
			n += nn
			if nn >= l:
				break
		return n


class LxTextReader(TextIOBase):

	def __init__(self, binr: RawIOBase, codeci: CodecInfo):
		self.idecoder = codeci.incrementaldecoder()
		self.binr = binr
		self._closed = False
		self.buf = bytearray(BUF_SZ)
		self.bmv = memoryview(self.buf)
		self.bsz = self._load_buf(self.bmv)
		self.str = self._next()
		self.slen = len(self.str)
		self.soffset = 0

	def _load_buf(self, buf: memoryview):
		n = self.binr.readinto(buf)
		if n is None or n <= 0:
			n = 0
			self.close()
		return n

	def _next(self):
		if self.bsz <= 0:
			return ""
		b = self.buf[:self.bsz]
		self.bsz = self._load_buf(self.bmv)
		final = self.bsz <= 0
		return self.idecoder.decode(b, final)

	def __enter__(self):
		return self

	def __del__(self):
		self.close()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.close()

	def close(self):
		if not self._closed:
			self._closed = True
			self.binr.close()

	@property
	def closed(self):
		return self._closed

	@property
	def encoding(self):
		return self._encoding

	def read(self, __size=...):
		if self._closed:
			return ""
		if __size is None:
			raise RuntimeError("Read all not allowed. Use size argument")
		if __size < 0:
			raise RuntimeError("Read all not allowed. Use size argument")
		if __size == 0:
			return ""
		if __size > self.bsz:
			raise RuntimeError(f"Can read at most {BUF_SZ} bytes")
		if self.soffset >= self.slen:
			self.str = self._next()
			if self.str == "":
				return ""
		res = ""
		sz = __size
		while sz > 0:
			len = self.slen - self.soffset
			if len > sz:
				last = self.soffset + sz
				res += self.str[self.soffset:last]
				self.soffset = last
				break
			else:
				res += self.str[self.soffset:]
				sz -= len
				self.str = self._next()
				if self.str == "":
					break
				self.slen = len(self.str)
				self.soffset = 0

		return res

	def detach(self):
		raise UnsupportedOperation("detach")

	def readline(self, __size=...):
		raise RuntimeError("Use read(sz)")

	def seek(self, __offset, __whence=...):
		raise OSError("Not seekable")

	def tell(self):
		raise OSError("Not seekable")

	def fileno(self):
		raise OSError("No fileno")

	def readlines(self, __hint=...):
		raise OSError("Use read(sz)")

	def readline(self, __size=...):
		raise OSError("Use read(sz)")

	def isatty(self):
		return False

	def flush(self) -> None:
		return None

	def readable(self):
		return True

	def seekable(self):
		return False

	def seek(self, __offset, __whence=...):
		raise OSError("Not seekable")

	def tell(self):
		raise OSError("Not seekable")

	def truncate(self, __size=...):
		raise OSError("Not seekable")

	def writable(self):
		return False

	def write(self, __b):
		raise OSError("Not writeable")

	def writelines(self, __lines):
		raise OSError("Not writeable")


class LxBlobReader(object):
	def __init__(self, cursor, bmsg):
		self.cursor = cursor
		self.first = bmsg
		self.last = None
		self.offset = 0
		self.len = 0
		self.current = None
		self._set_current(self.first)
		self.cursor.add_blob_reader(self)

	def _set_current(self, bmsg):
		self.last = bmsg
		self.offset = 0
		if self.last is None:
			self.len = 0
			self.current = None
		else:
			self.len = len(self.last.bytes_value)
			self.current = memoryview(self.last.bytes_value)

	def _next(self):
		if self.last.info.done:
			return False
		offset = self.last.info.offset + len(self.last.bytes_value)
		bmsg = self.cursor.read_blob(self.first.info.id, offset)
		if bmsg is None:
			return False
		self._set_current(bmsg)
		return True

	def readinto(self, buf):
		if buf is None:
			return 0
		blen = len(buf)
		if self.offset >= self.len:
			if not self._next():
				return 0
		n = 0
		while n < blen:
			bl = blen - n
			ll = self.len - self.offset
			if bl == ll:
				buf[n:] = self.current[self.offset:]
				n += bl
				self.offset += n
				break
			if bl < ll:
				inx = self.offset + bl
				buf[n:] = self.current[self.offset:inx]
				n += bl
				self.offset += n
				break
			inx = n + ll
			buf[n:inx] = self.current[self.offset:]
			n += ll
			self.offset += n
			if self.offset >= self.len:
				if not self._next():
					break
		return n

	def reset(self):
		self._set_current(self.first)

	def close(self, reset=True):
		if reset:
			self.reset()
		else:
			self.cursor = None
			self.first = None
			self.last = None
			self.offset = 0
			self.len = 0
			self.current = None


def typedValueToNative(v):
	if Rep.Name(v.type) == "BOOLEAN" or Rep.Name(v.type) == "PRIMITIVE_BOOLEAN":
		return v.bool_value

	elif Rep.Name(v.type) == "STRING" or Rep.Name(v.type) == "PRIMITIVE_CHAR" or Rep.Name(
			v.type) == "CHARACTER" or Rep.Name(v.type) == "BIG_DECIMAL":
		return v.string_value

	elif Rep.Name(v.type) == "FLOAT" or Rep.Name(v.type) == "PRIMITIVE_FLOAT" or Rep.Name(
			v.type) == "DOUBLE" or Rep.Name(v.type) == "PRIMITIVE_DOUBLE":
		return v.double_value

	elif Rep.Name(v.type) == "LONG" or Rep.Name(v.type) == "PRIMITIVE_LONG" or Rep.Name(
			v.type) == "INTEGER" or Rep.Name(v.type) == "PRIMITIVE_INT" or \
			Rep.Name(v.type) == "BIG_INTEGER" or Rep.Name(v.type) == "NUMBER" or Rep.Name(v.type) == "BYTE" or Rep.Name(
		v.type) == "PRIMITIVE_BYTE" or \
			Rep.Name(v.type) == "SHORT" or Rep.Name(v.type) == "PRIMITIVE_SHORT":
		return v.number_value

	elif Rep.Name(v.type) == "BYTE_STRING":
		return v.bytes_value

	else:
		return None


def javaTypetoNative(java_type, sql_type):
	if java_type == 'java.math.BigDecimal':
		return 'BIG_DECIMAL', None, "string_value", sql_type
	elif java_type == 'java.lang.Float':
		return 'FLOAT', float, "double_value", sql_type
	elif java_type == 'java.lang.Double':
		return 'DOUBLE', None, "double_value", sql_type
	elif java_type == 'java.lang.Long':
		return 'LONG', None, "number_value", sql_type
	elif java_type == 'java.lang.Integer':
		return 'INTEGER', int, "number_value", sql_type
	elif java_type == 'java.lang.Short':
		return 'SHORT', int, "number_value", sql_type
	elif java_type == 'java.lang.Byte':
		return 'BYTE', Binary, "bytes_value", sql_type
	elif java_type == 'java.lang.Boolean':
		return 'BOOLEAN', bool, "bool_value", sql_type
	elif java_type == 'java.lang.String':
		return 'STRING', None, "string_value", sql_type
	elif java_type == 'java.sql.Time':
		return 'JAVA_SQL_TIME', time_from_java_sql_time, "number_value", sql_type
	elif java_type == 'java.sql.Date':
		return 'JAVA_SQL_DATE', date_from_java_sql_date, "number_value", sql_type
	elif java_type == 'java.sql.Timestamp':
		return 'JAVA_SQL_TIMESTAMP', datetime_from_java_sql_timestamp, "number_value", sql_type
	elif java_type == '[B':
		return 'BYTE_STRING', bytes, "bytes_value", sql_type
	else:
		return 'NULL', None, None, sql_type


def nativeToParamType(value, isArray=False):
	if value is None:
		return 'NULL', None, "null"
	elif isinstance(value, int):
		if (abs(value) <= 0x7FFFFFFF and not isArray):
			return 'INTEGER', int, "number_value"
		else:
			return 'LONG', int, "number_value"
	elif isinstance(value, float):
		return 'DOUBLE', float, "double_value"
	elif isinstance(value, bool):
		return 'BOOLEAN', bool, "bool_value"
	elif isinstance(value, str):
		return 'STRING', str, "string_value"
	elif isinstance(value, datetime.time):
		return 'JAVA_SQL_TIME', time_to_java_sql_time, "number_value"
	elif isinstance(value,
					datetime.datetime):  # TODO check: important to put this case before datetime.date as datetimes also match in date clause
		return 'JAVA_SQL_TIMESTAMP', datetime_to_java_sql_timestamp, "number_value"
	elif isinstance(value, datetime.date):
		return 'JAVA_SQL_DATE', date_to_java_sql_date, "number_value"
	elif isinstance(value, bytes):
		return 'BYTE_STRING', bytes, "bytes_value"
	elif isinstance(value, LXARRAY):
		return 'ARRAY', None, "array_value"
	else:
		return 'STRING', str, "string_value"


def lxarray2tv(value: LXARRAY):
	result = common_pb2.TypedValue()  # parent typed value
	result.type = common_pb2.Rep.Value('ARRAY')
	if value.gettype() is int:
		result.component_type = common_pb2.Rep.Value("LONG")
	elif value.gettype() is float:
		result.component_type = common_pb2.Rep.Value("DOUBLE")
	elif value.gettype() is str:
		result.component_type = common_pb2.Rep.Value("STRING")
	else:
		raise ArrayTypeException("Array cannot be of type: {}".format(value.gettype()))

	for item in value.getvalue():
		type_param = nativeToParamType(item, isArray=True)  # int bigint varchar ???
		v_temp = common_pb2.TypedValue()
		v_temp.type = common_pb2.Rep.Value(type_param[0])
		if item != None:
			setattr(v_temp, type_param[2], type_param[1](item))
		else:
			setattr(v_temp, type_param[2], 1)  # 1 for True
		result.array_value.append(v_temp)
	return result


def typed_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	func = _rep_to_native_funcs[_rep_index_to_name[typed_value.type]]
	if func is None:
		_unsupported_typed_value(typed_value, ctx)
		return None
	return func(typed_value, ctx)


def _bool_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return bool(typed_value.bool_value)


def _string_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return str(typed_value.string_value)


def _big_decimal_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	value = str(typed_value.string_value)
	if '.' not in value and ',' not in value:
		return int(value)
	return Decimal(value)


def _bytes_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return bytes(typed_value.bytes_value)


def _number_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return int(typed_value.number_value)


def _double_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return float(typed_value.double_value)


def _null_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	return None


def _blob_value_2_native(typed_value: common_pb2.TypedValue, ctx):
	if typed_value.blob_info.done:
		return bytes(typed_value.bytes_value)
	bmsg = common_pb2.LxBlobMsg()
	bmsg.info.id = typed_value.blob_info.id
	bmsg.info.done = False
	bmsg.info.offset = 0
	bmsg.bytes_value = typed_value.bytes_value
	br = LxBlobReader(ctx, bmsg)
	return LxRawReader(br)


def _unsupported_typed_value(typed_value: common_pb2.TypedValue, ctx):
	raise NotImplementedError("Not supported: " + _rep_index_to_name[typed_value.type])


_rep_index_to_name = ['PRIMITIVE_BOOLEAN',
					  'PRIMITIVE_BYTE',
					  'PRIMITIVE_CHAR',
					  'PRIMITIVE_SHORT',
					  'PRIMITIVE_INT',
					  'PRIMITIVE_LONG',
					  'PRIMITIVE_FLOAT',
					  'PRIMITIVE_DOUBLE',
					  'BOOLEAN',
					  'BYTE',
					  'CHARACTER',
					  'SHORT',
					  'INTEGER',
					  'LONG',
					  'FLOAT',
					  'DOUBLE',
					  'JAVA_SQL_TIME',
					  'JAVA_SQL_TIMESTAMP',
					  'JAVA_SQL_DATE',
					  'JAVA_UTIL_DATE',
					  'BYTE_STRING',
					  'STRING',
					  'NUMBER',
					  'OBJECT',
					  'NULL',
					  'BIG_INTEGER',
					  'BIG_DECIMAL',
					  'ARRAY',
					  'STRUCT',
					  'MULTISET',
					  'ZONED_DATE_TIME',
					  'LXBLOB']

_rep_to_native_funcs = {'ARRAY': _unsupported_typed_value,
						'BIG_DECIMAL': _big_decimal_value_2_native,
						'BIG_INTEGER': _bytes_value_2_native,
						'BOOLEAN': _bool_value_2_native,
						'BYTE': _number_value_2_native,
						'BYTE_STRING': _bytes_value_2_native,
						'CHARACTER': _string_value_2_native,
						'DOUBLE': _double_value_2_native,
						'FLOAT': _double_value_2_native,
						'INTEGER': _number_value_2_native,
						'JAVA_SQL_DATE': _number_value_2_native,
						'JAVA_SQL_TIME': _number_value_2_native,
						'JAVA_SQL_TIMESTAMP': _number_value_2_native,
						'JAVA_UTIL_DATE': _number_value_2_native,
						'LONG': _number_value_2_native,
						'MULTISET': _unsupported_typed_value,
						'NULL': _null_value_2_native,
						'NUMBER': _number_value_2_native,
						'OBJECT': _unsupported_typed_value,
						'PRIMITIVE_BOOLEAN': _bool_value_2_native,
						'PRIMITIVE_BYTE': _number_value_2_native,
						'PRIMITIVE_CHAR': _string_value_2_native,
						'PRIMITIVE_DOUBLE': _number_value_2_native,
						'PRIMITIVE_FLOAT': _double_value_2_native,
						'PRIMITIVE_INT': _number_value_2_native,
						'PRIMITIVE_LONG': _number_value_2_native,
						'PRIMITIVE_SHORT': _number_value_2_native,
						'SHORT': _number_value_2_native,
						'STRING': _string_value_2_native,
						'STRUCT': _unsupported_typed_value,
						'LXBLOB': _blob_value_2_native}


def lxarray2tv(value: LXARRAY):
	result = common_pb2.TypedValue()  # parent typed value
	result.type = common_pb2.Rep.Value('ARRAY')
	if value.gettype() is int:
		result.component_type = common_pb2.Rep.Value("LONG")
	elif value.gettype() is float:
		result.component_type = common_pb2.Rep.Value("DOUBLE")
	elif value.gettype() is str:
		result.component_type = common_pb2.Rep.Value("STRING")
	else:
		raise ArrayTypeException("Array cannot be of type: {}".format(value.gettype()))

	for item in value.getvalue():
		type_param = nativeToParamType(item, isArray=True)  # int bigint varchar ???
		v_temp = common_pb2.TypedValue()
		v_temp.type = common_pb2.Rep.Value(type_param[0])
		if item != None:
			setattr(v_temp, type_param[2], type_param[1](item))
		else:
			setattr(v_temp, type_param[2], 1)  # 1 for True
		result.array_value.append(v_temp)
	return result


def blob2tv(value: IOBase, ctx):
	txt2bin = None
	if isinstance(value, TextIOBase):
		txt2bin = LxText2Bin(value, ctx._connection.codeci)
		lxblob = LxRawReader(txt2bin)
	elif isinstance(value, LxRawReader):
		lxblob = value
	else:
		lxblob = LxRawReader(value)

	result = common_pb2.TypedValue()
	bs = lxblob.read(BUF_SZ)
	if lxblob.closed:
		if txt2bin is None:
			result.bytes_value = bs
			result.type = common_pb2.Rep.Value('BYTE_STRING')
			return result
		result.type = common_pb2.Rep.Value('STRING')
		result.string_value = txt2bin.first
		return result
	result.type = common_pb2.Rep.Value('LXBLOB')
	result.bytes_value = bs
	bid = ctx.add_blob(lxblob)
	result.blob_info.id = bid
	result.blob_info.offset = 0
	result.blob_info.done = False
	return result


def native_2_typed_value(value, ctx):
	if value is None:
		return common_pb2.TypedValue(null=True, type=common_pb2.Rep.Value('NULL'))
	if isinstance(value, LXARRAY):
		return lxarray2tv(value)
	if isinstance(value, IOBase):
		return blob2tv(value, ctx)

	type_param = nativeToParamType(value)
	styp = type_param[0]
	result = common_pb2.TypedValue()
	result.type = common_pb2.Rep.Value(styp)
	if type_param[1] is not None:
		if isinstance(value, datetime.datetime):
			if value.tzinfo is None:
				value = value.astimezone(None)
		setattr(result, type_param[2], type_param[1](value))
		return result
	setattr(result, type_param[2], value)
	return result



# Copyright 2017 Dimitri Capitaine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ._vendor.future.standard_library import install_aliases
install_aliases()

import urllib.parse
import lxdbapi

from sqlalchemy.engine.default import DefaultDialect

from sqlalchemy.sql.compiler import DDLCompiler, SQLCompiler
from sqlalchemy.sql import compiler
from sqlalchemy.engine import default

from sqlalchemy import types, BigInteger, Float
from sqlalchemy.types import INTEGER, BIGINT, SMALLINT, VARCHAR, CHAR, \
    FLOAT, DATE, BOOLEAN, DECIMAL, TIMESTAMP, TIME, VARBINARY


class LeanxcaleCompiler(SQLCompiler):
    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleDDLCompiler(DDLCompiler):

    def visit_primary_key_constraint(self, constraint):
        # if constraint.name is None:
        #    raise CompileError("can't create primary key without a name")
        return DDLCompiler.visit_primary_key_constraint(self, constraint)

    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleExecutionContext(default.DefaultExecutionContext):

    def should_autocommit_text(self, statement):
        pass

    def create_server_side_cursor(self):
        pass

    def fire_sequence(self, seq, type_):
        return self._execute_scalar(
            (
                    "SELECT NEXT VALUE FOR %s"
                    % seq.name
            ),
            type_
        )


class LeanXcaleIdentifierPreparer(compiler.IdentifierPreparer):

    def __init__(self, dialect, server_ansiquotes=False, **kw):

        quote = '"'

        super(LeanXcaleIdentifierPreparer, self).__init__(
            dialect, initial_quote=quote, escape_quote=quote
        )

    def _quote_free_identifiers(self, *ids):
        """Unilaterally identifier-quote any number of strings."""

        return tuple([self.quote_identifier(i) for i in ids if i is not None])

class LeanXcaleGenericTypeCompiler(compiler.GenericTypeCompiler):
    def visit_ARRAY(self, type_, **kw):
        if type_.item_type.python_type == int:
            return "BIGINT ARRAY"
        elif type_.item_type.python_type == float:
            return "DOUBLE ARRAY"
        elif type_.item_type.python_type == str:
            return "VARCHAR ARRAY"
        else:
            raise Exception("ARRAY of type {} is not supported".format(str(type_)))


class LeanxcaleDialect(DefaultDialect):
    name = "leanxcale"

    driver = "lxdbapi"

    ddl_compiler = LeanxcaleDDLCompiler
    preparer = LeanXcaleIdentifierPreparer
    statement_compiler = LeanxcaleCompiler
    type_compiler = LeanXcaleGenericTypeCompiler

    supports_sequences = True
    sequences_optional = True
    supports_multivalues_insert = True

    execution_ctx_cls = LeanxcaleExecutionContext

    preexecute_autoincrement_sequences = True

    @classmethod
    def dbapi(cls):

        return lxdbapi

    def create_connect_args(self, url):

        # Deal with PARALLEL option
        parallel = url.query.get('parallel')

        # Deal with SECURE option
        secure = url.query.get('secure')

        distribute = url.query.get('distribute')

        appTimeZone = url.query.get('appTimeZone')

        schname = url.query.get('schema')

        leanxcale_url = urllib.parse.urlunsplit(urllib.parse.SplitResult(
            scheme='http',
            netloc='{}:{}'.format(url.host, url.port or 8765),
            path='/',
            query=urllib.parse.urlencode(url.query),
            fragment='',
        ))
        if url.query.get('autocommit') == 'False':
            autocommit = False
        else:
            autocommit = True

        params = {'autocommit': autocommit, 'user': url.username}
        if parallel is not None:
            params.update({'parallel': parallel})
        if secure is not None:
            params.update({'secure': secure})
        if distribute is not None:
            params.update({'distribute': distribute})
        if url.password is not None:
            params.update({'password': url.password})
        if appTimeZone is not None:
            params.update({'appTimeZone': appTimeZone})
        if schname is not None:
            params.update({'schema': schname})

        params.update({'database': url.database})
        return [leanxcale_url], params

    def do_rollback(self, dbapi_conection):
        dbapi_conection.rollback()

    def do_commit(self, dbapi_conection):
        dbapi_conection.commit()

    def has_sequence(self, connection, sequence_name, schema=None, **kw):
        return sequence_name in self.get_sequence_names(connection, schema)

    def get_sequence_names(self, connection, schema=None, **kw):
        schema = schema if schema is not None else self.default_schema_name
        params = []
        sql = "SELECT DISTINCT(tableName) from LXSYS.TABLES where tableSchem <> 'LXSYS'"
        sql += " AND tableType = 'SEQUENCE'"
        if schema:
            sql += " AND tableSchem = '?'"
            params.append(str(schema))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(row[0])
            row = cursor.fetchone()
        return result

    def has_table(self, connection, table_name, schema=None, **kw):
        return table_name in self.get_table_names(connection, schema)

    def _get_default_schema_name(self, connection):
        sql = "select schema from LXSYS.CONNECTIONS where iscurrent"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql)
        row = cursor.fetchone()
        if row:
            return row[0]
        return None

    def get_schema_names(self, connection, **kw):
        sql = "select distinct(tableSchem) from LXSYS.TABLES where tableSchem <> 'LXSYS'"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql)
        result = []
        schema = cursor.fetchone()
        while schema is not None:
            result.append(str(schema[0]))
            schema = cursor.fetchone()
        return result

    def get_table_names(self, connection, schema=None, **kw):
        schema = schema if schema is not None else self.default_schema_name
        params = []
        sql = "SELECT DISTINCT(tableName) from LXSYS.TABLES where tableSchem <> 'LXSYS'"
        sql += " AND tableType IN ('TABLE', 'MATERIALIZED QUERY TABLE')"
        if schema:
            sql += " AND tableSchem = ?"
            params.append(str(schema))
        if 'tableNamePattern' in kw.keys():
            tableNamePattern = kw.get('tableNamePattern')
            sql += " AND tableName LIKE ?"
            params.append(str(tableNamePattern))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(str(row[0]))
            row = cursor.fetchone()
        return result

    def get_columns(self, connection, table_name=None, schema=None, catalog=None, column_name=None, **kw):
        params = []
        sql = "select columnName, dataType, typeName, nullable, columnDef from LXSYS.COLUMNS"
        schema = schema if schema is not None else self.default_schema_name
        has_where = False
        if catalog:
            sql += " WHERE tableCat = ?"
            has_where = True
            params.append(str(catalog))
        if schema:
            if has_where:
                sql += " AND "
            else:
                sql += " WHERE "
                has_where = True
            sql += "tableSchem = ?"
            params.append(str(schema))
        if table_name:
            if has_where:
                sql += " AND "
            else:
                sql += " WHERE "
                has_where = True
            sql += "tableName = ?"
            params.append(str(table_name))
        if column_name:
            if has_where:
                sql += " AND "
            else:
                sql += " WHERE "
            sql += "columnName = ?"
            params.append(str(column_name))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            col_d = {}
            col_d.update({'name': row[0]})
            if row[1] == 2003:
                if row[2].split()[0] == 'BIGINT':
                    col_d.update({'item_type': BigInteger})
                elif row[2].split()[0] == 'DOUBLE':
                    col_d.update({'item_type': Float})
                elif row[2].split()[0] == 'VARCHAR':
                    col_d.update({'item_type': VARCHAR})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == 2104:
                col_d.update({'item_type': BigInteger})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == 2108:
                col_d.update({'item_type': Float})
                col_d.update({'type': ARRAY(row[5])})
            elif row[1] == 2112:
                col_d.update({'item_type': VARCHAR})
                col_d.update({'type': ARRAY(row[5])})
            else:
                col_d.update({'type': COLUMN_DATA_TYPE[row[1]]})
            col_d.update({'nullable': row[3] == 1 if True else False})
            col_d.update({'default': row[4]})
            result.append(col_d)
            row = cursor.fetchone()

        return result

    def get_view_names(self, connection, schema=None, **kw):
        params = []
        schema = schema if schema is not None else self.default_schema_name
        sql = "SELECT DISTINCT(tableName) from LXSYS.TABLES where tableSchem <> 'LXSYS'"
        sql += " AND tableType = 'VIEW'"
        if schema:
            sql += " AND tableSchem = ?"
            params.append(str(schema))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        row = cursor.fetchone()
        while row is not None:
            result.append(row[0])
            row = cursor.fetchone()
        return result

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT columnName FROM LXSYS.PRIMARY_KEYS WHERE TABLENAME = ? ORDER BY keyseq"
        params.append(str(table_name))
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        cols = []
        row = cursor.fetchone()
        while row is not None:
            cols.append(row[0])
            row = cursor.fetchone()

        result = {
            'constrained_columns': cols,
        }
        return result

#   TABLE_CAT   TABLE_SCHEM   TABLE_NAME   COLUMN_NAME   DATA_TYPE   TYPE_NAME   COLUMN_SIZE   BUFFER_LENGTH   DECIMAL_DIGITS   NUM_PREC_RADIX   NULLABLE   REMARKS   COLUMN_DEF   SQL_DATA_TYPE   SQL_DATETIME_SUB   CHAR_OCTET_LENGTH   ORDINAL_POSITION   IS_NULLABLE   SCOPE_CATALOG   SCOPE_SCHEMA   SCOPE_TABLE   SOURCE_DATA_TYPE   IS_AUTOINCREMENT   IS_GENERATEDCOLUMN  

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT fktableSchem, fktableName, fkName, pktableSchem, pktableName, pkcolumnName, fkcolumnName, keySeq"\
              " from LXSYS.FOREIGN_KEYS" \
              " WHERE fktableName = ?"
        params.append(str(table_name))
        if schema:
            sql += " and fktableSchem = ?"
            params.append(str(schema))
        sql += " order by fktableSchem, fktableName, fkName, keySeq"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        fksch = None
        fktbl = None
        fkname = None
        pksch = None
        pktbl = None
        fkcols = None
        pkcols = None
        row = cursor.fetchone()
        while row is not None:
            if fkname != row[2] or fktbl != row[1] or fksch != row[0]:
                if fkname:
                    fk_info = {
                        "constrained_columns": fkcols,
                        "referred_schema": pksch,
                        "referred_table": pktbl,
                        "referred_columns": pkcols,
                        "name": fkname,
                    }
                    result.append(fk_info)
                fksch = row[0]
                fktbl = row[1]
                fkname = row[2]
                pksch = row[3]
                pktbl = row[4]
                fkcols = []
                pkcols = []
            pkcols.append(row[5])
            fkcols.append(row[6])
            row = cursor.fetchone()

        if fkname:
            fk_info = {
                "constrained_columns": fkcols,
                "referred_schema": pksch,
                "referred_table": pktbl,
                "referred_columns": pkcols,
                "name": fkname,
            }
            result.append(fk_info)
        return result

    def get_indexes(self, connection, table_name, schema=None, **kw):
        params = []
        sql = "SELECT indexName, columnName, cast(nonUnique as INTEGER) FROM LXSYS.INDEX_COLUMNS WHERE TABLENAME = ?"
        params.append(str(table_name))
        if schema:
            sql += " and tableSchem = ?"
            params.append(str(schema))
        sql += " order by indexname, ordinalposition"
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute(sql, params)
        result = []
        name = None
        cols = None
        nonUnique = None
        row = cursor.fetchone()
        while row is not None:
            if name != row[0]:
                if name is not None:
                    item = {
                        'name': name,
                        'column_names': cols,
                        'unique': not nonUnique
                    }
                    result.append(item)
                name = row[0]
                cols = []
                if row[2] is None or int(row[2]) == 0:
                    nonUnique = False
                else:
                    nonUnique = True
            cols.append(row[1])
            row = cursor.fetchone()

        if name is not None:
            item = {
                'name': name,
                'column_names': cols,
                'unique': not nonUnique
            }
            result.append(item)
        return result

class TINYINT(types.Integer):
    __visit_name__ = "INTEGER"


class UTINYINT(types.Integer):
    __visit_name__ = "INTEGER"


class UINTEGER(types.Integer):
    __visit_name__ = "INTEGER"


class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class UDOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class UFLOAT(types.FLOAT):
    __visit_name__ = "FLOAT"


class ULONG(types.BIGINT):
    __visit_name__ = "BIGINT"


class UTIME(types.TIME):
    __visit_name__ = "TIME"


class UDATE(types.DATE):
    __visit_name__ = "DATE"


class UTIMESTAMP(types.TIMESTAMP):
    __visit_name__ = "TIMESTAMP"


class ROWID(types.String):
    __visit_name__ = "VARCHAR"

class ARRAY(types.ARRAY):
    __visit_name__ = "ARRAY"
    def __init__(self, type):
        types.ARRAY.__init__(self, type)



COLUMN_DATA_TYPE = {
    -6: TINYINT,
    -5: BIGINT,
    -3: VARBINARY,
    1: CHAR,
    3: DECIMAL,
    4: INTEGER,
    5: SMALLINT,
    6: FLOAT,
    8: DOUBLE,
    9: UINTEGER,
    10: ULONG,
    11: UTINYINT,
    12: VARCHAR,
    13: ROWID,
    14: UFLOAT,
    15: UDOUBLE,
    16: BOOLEAN,
    18: UTIME,
    19: UDATE,
    20: UTIMESTAMP,
    91: DATE,
    92: TIME,
    93: TIMESTAMP
}
